# Lint as: python3
# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of FlexBuffers binary format.

For more info check https://google.github.io/flatbuffers/flexbuffers.html and
corresponding C++ implementation at
https://github.com/google/flatbuffers/blob/master/include/flatbuffers/flexbuffers.h
"""

# pylint: disable=invalid-name
# TODO(dkovalev): Add type hints everywhere, so tools like pytypes could work.

import array
import contextlib
import enum
import struct

__all__ = ("Type", "Builder", "GetRoot", "Dumps", "Loads")


class BitWidth(enum.IntEnum):
    """Supported bit widths of value types.

    These are used in the lower 2 bits of a type field to determine the size of
    the elements (and or size field) of the item pointed to (e.g. vector).
    """

    W8 = 0  # 2^0 = 1 byte
    W16 = 1  # 2^1 = 2 bytes
    W32 = 2  # 2^2 = 4 bytes
    W64 = 3  # 2^3 = 8 bytes

    @staticmethod
    def U(value):
        """Returns the minimum `BitWidth` to encode unsigned integer value."""
        assert value >= 0

        if value < (1 << 8):
            return BitWidth.W8
        elif value < (1 << 16):
            return BitWidth.W16
        elif value < (1 << 32):
            return BitWidth.W32
        elif value < (1 << 64):
            return BitWidth.W64
        else:
            raise ValueError("value is too big to encode: %s" % value)

    @staticmethod
    def I(value):
        """Returns the minimum `BitWidth` to encode signed integer value."""
        # -2^(n-1) <=     value < 2^(n-1)
        # -2^n     <= 2 * value < 2^n
        # 2 * value < 2^n, when value >= 0 or 2 * (-value) <= 2^n, when value < 0
        # 2 * value < 2^n, when value >= 0 or 2 * (-value) - 1 < 2^n, when value < 0
        #
        # if value >= 0:
        #   return BitWidth.U(2 * value)
        # else:
        #   return BitWidth.U(2 * (-value) - 1)  # ~x = -x - 1
        value *= 2
        return BitWidth.U(value if value >= 0 else ~value)

    @staticmethod
    def F(value):
        """Returns the `BitWidth` to encode floating point value."""
        if struct.unpack("f", struct.pack("f", value))[0] == value:
            return BitWidth.W32
        return BitWidth.W64

    @staticmethod
    def B(byte_width):
        return {1: BitWidth.W8, 2: BitWidth.W16, 4: BitWidth.W32, 8: BitWidth.W64}[
            byte_width
        ]


I = {1: "b", 2: "h", 4: "i", 8: "q"}  # Integer formats
U = {1: "B", 2: "H", 4: "I", 8: "Q"}  # Unsigned integer formats
F = {4: "f", 8: "d"}  # Floating point formats


def _Unpack(fmt, buf):
    return struct.unpack(fmt[len(buf)], buf)[0]


def _UnpackVector(fmt, buf, length):
    byte_width = len(buf) // length
    return struct.unpack("%d%s" % (length, fmt[byte_width]), buf)


def _Pack(fmt, value, byte_width):
    return struct.pack(fmt[byte_width], value)


def _PackVector(fmt, values, byte_width):
    return struct.pack("%d%s" % (len(values), fmt[byte_width]), *values)


def _Mutate(fmt, buf, value, byte_width, value_bit_width):
    if (1 << value_bit_width) <= byte_width:
        buf[:byte_width] = _Pack(fmt, value, byte_width)
        return True
    return False


# Computes how many bytes you'd have to pad to be able to write an
# "scalar_size" scalar if the buffer had grown to "buf_size",
# "scalar_size" is a power of two.
def _PaddingBytes(buf_size, scalar_size):
    # ((buf_size + (scalar_size - 1)) // scalar_size) * scalar_size - buf_size
    return -buf_size & (scalar_size - 1)


def _ShiftSlice(s, offset, length):
    start = offset + (0 if s.start is None else s.start)
    stop = offset + (length if s.stop is None else s.stop)
    return slice(start, stop, s.step)


# https://en.cppreference.com/w/cpp/algorithm/lower_bound
def _LowerBound(values, value, pred):
    """Implementation of C++ std::lower_bound() algorithm."""
    first, last = 0, len(values)
    count = last - first
    while count > 0:
        i = first
        step = count // 2
        i += step
        if pred(values[i], value):
            i += 1
            first = i
            count -= step + 1
        else:
            count = step
    return first


# https://en.cppreference.com/w/cpp/algorithm/binary_search
def _BinarySearch(values, value, pred=lambda x, y: x < y):
    """Implementation of C++ std::binary_search() algorithm."""
    index = _LowerBound(values, value, pred)
    if index != len(values) and not pred(value, values[index]):
        return index
    return -1


class Type(enum.IntEnum):
    """Supported types of encoded data.

    These are used as the upper 6 bits of a type field to indicate the actual
    type.
    """

    NULL = 0
    INT = 1
    UINT = 2
    FLOAT = 3
    # Types above stored inline, types below store an offset.
    KEY = 4
    STRING = 5
    INDIRECT_INT = 6
    INDIRECT_UINT = 7
    INDIRECT_FLOAT = 8
    MAP = 9
    VECTOR = 10  # Untyped.

    VECTOR_INT = 11  # Typed any size (stores no type table).
    VECTOR_UINT = 12
    VECTOR_FLOAT = 13
    VECTOR_KEY = 14
    # DEPRECATED, use VECTOR or VECTOR_KEY instead.
    # Read test.cpp/FlexBuffersDeprecatedTest() for details on why.
    VECTOR_STRING_DEPRECATED = 15

    VECTOR_INT2 = 16  # Typed tuple (no type table, no size field).
    VECTOR_UINT2 = 17
    VECTOR_FLOAT2 = 18
    VECTOR_INT3 = 19  # Typed triple (no type table, no size field).
    VECTOR_UINT3 = 20
    VECTOR_FLOAT3 = 21
    VECTOR_INT4 = 22  # Typed quad (no type table, no size field).
    VECTOR_UINT4 = 23
    VECTOR_FLOAT4 = 24

    BLOB = 25
    BOOL = 26
    VECTOR_BOOL = 36  # To do the same type of conversion of type to vector type

    @staticmethod
    def Pack(type_, bit_width):
        return (int(type_) << 2) | bit_width

    @staticmethod
    def Unpack(packed_type):
        return 1 << (packed_type & 0b11), Type(packed_type >> 2)

    @staticmethod
    def IsInline(type_):
        return type_ <= Type.FLOAT or type_ == Type.BOOL

    @staticmethod
    def IsTypedVector(type_):
        return (
            Type.VECTOR_INT <= type_ <= Type.VECTOR_STRING_DEPRECATED
            or type_ == Type.VECTOR_BOOL
        )

    @staticmethod
    def IsTypedVectorElementType(type_):
        return Type.INT <= type_ <= Type.STRING or type_ == Type.BOOL

    @staticmethod
    def ToTypedVectorElementType(type_):
        if not Type.IsTypedVector(type_):
            raise ValueError("must be typed vector type")

        return Type(type_ - Type.VECTOR_INT + Type.INT)

    @staticmethod
    def IsFixedTypedVector(type_):
        return Type.VECTOR_INT2 <= type_ <= Type.VECTOR_FLOAT4

    @staticmethod
    def IsFixedTypedVectorElementType(type_):
        return Type.INT <= type_ <= Type.FLOAT

    @staticmethod
    def ToFixedTypedVectorElementType(type_):
        if not Type.IsFixedTypedVector(type_):
            raise ValueError("must be fixed typed vector type")

        # 3 types each, starting from length 2.
        fixed_type = type_ - Type.VECTOR_INT2
        return Type(fixed_type % 3 + Type.INT), fixed_type // 3 + 2

    @staticmethod
    def ToTypedVector(element_type, fixed_len=0):
        """Converts element type to corresponding vector type.

        Args:
          element_type: vector element type
          fixed_len: number of elements: 0 for typed vector; 2, 3, or 4 for fixed
            typed vector.

        Returns:
          Typed vector type or fixed typed vector type.
        """
        if fixed_len == 0:
            if not Type.IsTypedVectorElementType(element_type):
                raise ValueError("must be typed vector element type")
        else:
            if not Type.IsFixedTypedVectorElementType(element_type):
                raise ValueError("must be fixed typed vector element type")

        offset = element_type - Type.INT
        if fixed_len == 0:
            return Type(offset + Type.VECTOR_INT)  # TypedVector
        elif fixed_len == 2:
            return Type(offset + Type.VECTOR_INT2)  # FixedTypedVector
        elif fixed_len == 3:
            return Type(offset + Type.VECTOR_INT3)  # FixedTypedVector
        elif fixed_len == 4:
            return Type(offset + Type.VECTOR_INT4)  # FixedTypedVector
        else:
            raise ValueError("unsupported fixed_len: %s" % fixed_len)


class Buf:
    """Class to access underlying buffer object starting from the given offset."""

    def __init__(self, buf, offset):
        self._buf = buf
        self._offset = offset if offset >= 0 else len(buf) + offset
        self._length = len(buf) - self._offset

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self._buf[_ShiftSlice(key, self._offset, self._length)]
        elif isinstance(key, int):
            return self._buf[self._offset + key]
        else:
            raise TypeError("invalid key type")

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            self._buf[_ShiftSlice(key, self._offset, self._length)] = value
        elif isinstance(key, int):
            self._buf[self._offset + key] = key
        else:
            raise TypeError("invalid key type")

    def __repr__(self):
        return "buf[%d:]" % self._offset

    def Find(self, sub):
        """Returns the lowest index where the sub subsequence is found."""
        return self._buf[self._offset :].find(sub)

    def Slice(self, offset):
        """Returns new `Buf` which starts from the given offset."""
        return Buf(self._buf, self._offset + offset)

    def Indirect(self, offset, byte_width):
        """Return new `Buf` based on the encoded offset (indirect encoding)."""
        return self.Slice(offset - _Unpack(U, self[offset : offset + byte_width]))


class Object:
    """Base class for all non-trivial data accessors."""

    __slots__ = "_buf", "_byte_width"

    def __init__(self, buf, byte_width):
        self._buf = buf
        self._byte_width = byte_width

    @property
    def ByteWidth(self):
        return self._byte_width


class Sized(Object):
    """Base class for all data accessors which need to read encoded size."""

    __slots__ = ("_size",)

    def __init__(self, buf, byte_width, size=0):
        super().__init__(buf, byte_width)
        if size == 0:
            self._size = _Unpack(U, self.SizeBytes)
        else:
            self._size = size

    @property
    def SizeBytes(self):
        return self._buf[-self._byte_width : 0]

    def __len__(self):
        return self._size


class Blob(Sized):
    """Data accessor for the encoded blob bytes."""

    __slots__ = ()

    @property
    def Bytes(self):
        return self._buf[0 : len(self)]

    def __repr__(self):
        return "Blob(%s, size=%d)" % (self._buf, len(self))


class String(Sized):
    """Data accessor for the encoded string bytes."""

    __slots__ = ()

    @property
    def Bytes(self):
        return self._buf[0 : len(self)]

    def Mutate(self, value):
        """Mutates underlying string bytes in place.

        Args:
          value: New string to replace the existing one. New string must have less
            or equal UTF-8-encoded bytes than the existing one to successfully
            mutate underlying byte buffer.

        Returns:
          Whether the value was mutated or not.
        """
        encoded = value.encode("utf-8")
        n = len(encoded)
        if n <= len(self):
            self._buf[-self._byte_width : 0] = _Pack(U, n, self._byte_width)
            self._buf[0:n] = encoded
            self._buf[n : len(self)] = bytearray(len(self) - n)
            return True
        return False

    def __str__(self):
        return self.Bytes.decode("utf-8")

    def __repr__(self):
        return "String(%s, size=%d)" % (self._buf, len(self))


class Key(Object):
    """Data accessor for the encoded key bytes."""

    __slots__ = ()

    def __init__(self, buf, byte_width):
        assert byte_width == 1
        super().__init__(buf, byte_width)

    @property
    def Bytes(self):
        return self._buf[0 : len(self)]

    def __len__(self):
        return self._buf.Find(0)

    def __str__(self):
        return self.Bytes.decode("ascii")

    def __repr__(self):
        return "Key(%s, size=%d)" % (self._buf, len(self))


class Vector(Sized):
    """Data accessor for the encoded vector bytes."""

    __slots__ = ()

    def __getitem__(self, index):
        if index < 0 or index >= len(self):
            raise IndexError(
                "vector index %s is out of [0, %d) range" % (index, len(self))
            )

        packed_type = self._buf[len(self) * self._byte_width + index]
        buf = self._buf.Slice(index * self._byte_width)
        return Ref.PackedType(buf, self._byte_width, packed_type)

    @property
    def Value(self):
        """Returns the underlying encoded data as a list object."""
        return [e.Value for e in self]

    def __repr__(self):
        return "Vector(%s, byte_width=%d, size=%d)" % (
            self._buf,
            self._byte_width,
            self._size,
        )


class TypedVector(Sized):
    """Data accessor for the encoded typed vector or fixed typed vector bytes."""

    __slots__ = "_element_type", "_size"

    def __init__(self, buf, byte_width, element_type, size=0):
        super().__init__(buf, byte_width, size)

        if element_type == Type.STRING:
            # These can't be accessed as strings, since we don't know the bit-width
            # of the size field, see the declaration of
            # FBT_VECTOR_STRING_DEPRECATED above for details.
            # We change the type here to be keys, which are a subtype of strings,
            # and will ignore the size field. This will truncate strings with
            # embedded nulls.
            element_type = Type.KEY

        self._element_type = element_type

    @property
    def Bytes(self):
        return self._buf[: self._byte_width * len(self)]

    @property
    def ElementType(self):
        return self._element_type

    def __getitem__(self, index):
        if index < 0 or index >= len(self):
            raise IndexError(
                "vector index %s is out of [0, %d) range" % (index, len(self))
            )

        buf = self._buf.Slice(index * self._byte_width)
        return Ref(buf, self._byte_width, 1, self._element_type)

    @property
    def Value(self):
        """Returns underlying data as list object."""
        if not self:
            return []

        if self._element_type is Type.BOOL:
            return [bool(e) for e in _UnpackVector(U, self.Bytes, len(self))]
        elif self._element_type is Type.INT:
            return list(_UnpackVector(I, self.Bytes, len(self)))
        elif self._element_type is Type.UINT:
            return list(_UnpackVector(U, self.Bytes, len(self)))
        elif self._element_type is Type.FLOAT:
            return list(_UnpackVector(F, self.Bytes, len(self)))
        elif self._element_type is Type.KEY:
            return [e.AsKey for e in self]
        elif self._element_type is Type.STRING:
            return [e.AsString for e in self]
        else:
            raise TypeError("unsupported element_type: %s" % self._element_type)

    def __repr__(self):
        return "TypedVector(%s, byte_width=%d, element_type=%s, size=%d)" % (
            self._buf,
            self._byte_width,
            self._element_type,
            self._size,
        )


class Map(Vector):
    """Data accessor for the encoded map bytes."""

    @staticmethod
    def CompareKeys(a, b):
        if isinstance(a, Ref):
            a = a.AsKeyBytes
        if isinstance(b, Ref):
            b = b.AsKeyBytes
        return a < b

    def __getitem__(self, key):
        if isinstance(key, int):
            return super().__getitem__(key)

        index = _BinarySearch(self.Keys, key.encode("ascii"), self.CompareKeys)
        if index != -1:
            return super().__getitem__(index)

        raise KeyError(key)

    @property
    def Keys(self):
        byte_width = _Unpack(U, self._buf[-2 * self._byte_width : -self._byte_width])
        buf = self._buf.Indirect(-3 * self._byte_width, self._byte_width)
        return TypedVector(buf, byte_width, Type.KEY)

    @property
    def Values(self):
        return Vector(self._buf, self._byte_width)

    @property
    def Value(self):
        return {k.Value: v.Value for k, v in zip(self.Keys, self.Values)}

    def __repr__(self):
        return "Map(%s, size=%d)" % (self._buf, len(self))


class Ref:
    """Data accessor for the encoded data bytes."""

    __slots__ = "_buf", "_parent_width", "_byte_width", "_type"

    @staticmethod
    def PackedType(buf, parent_width, packed_type):
        byte_width, type_ = Type.Unpack(packed_type)
        return Ref(buf, parent_width, byte_width, type_)

    def __init__(self, buf, parent_width, byte_width, type_):
        self._buf = buf
        self._parent_width = parent_width
        self._byte_width = byte_width
        self._type = type_

    def __repr__(self):
        return "Ref(%s, parent_width=%d, byte_width=%d, type_=%s)" % (
            self._buf,
            self._parent_width,
            self._byte_width,
            self._type,
        )

    @property
    def _Bytes(self):
        return self._buf[: self._parent_width]

    def _ConvertError(self, target_type):
        raise TypeError("cannot convert %s to %s" % (self._type, target_type))

    def _Indirect(self):
        return self._buf.Indirect(0, self._parent_width)

    @property
    def IsNull(self):
        return self._type is Type.NULL

    @property
    def IsBool(self):
        return self._type is Type.BOOL

    @property
    def AsBool(self):
        if self._type is Type.BOOL:
            return bool(_Unpack(U, self._Bytes))
        else:
            return self.AsInt != 0

    def MutateBool(self, value):
        """Mutates underlying boolean value bytes in place.

        Args:
          value: New boolean value.

        Returns:
          Whether the value was mutated or not.
        """
        return self.IsBool and _Mutate(
            U, self._buf, value, self._parent_width, BitWidth.W8
        )

    @property
    def IsNumeric(self):
        return self.IsInt or self.IsFloat

    @property
    def IsInt(self):
        return self._type in (
            Type.INT,
            Type.INDIRECT_INT,
            Type.UINT,
            Type.INDIRECT_UINT,
        )

    @property
    def AsInt(self):
        """Returns current reference as integer value."""
        if self.IsNull:
            return 0
        elif self.IsBool:
            return int(self.AsBool)
        elif self._type is Type.INT:
            return _Unpack(I, self._Bytes)
        elif self._type is Type.INDIRECT_INT:
            return _Unpack(I, self._Indirect()[: self._byte_width])
        if self._type is Type.UINT:
            return _Unpack(U, self._Bytes)
        elif self._type is Type.INDIRECT_UINT:
            return _Unpack(U, self._Indirect()[: self._byte_width])
        elif self.IsString:
            return len(self.AsString)
        elif self.IsKey:
            return len(self.AsKey)
        elif self.IsBlob:
            return len(self.AsBlob)
        elif self.IsVector:
            return len(self.AsVector)
        elif self.IsTypedVector:
            return len(self.AsTypedVector)
        elif self.IsFixedTypedVector:
            return len(self.AsFixedTypedVector)
        else:
            raise self._ConvertError(Type.INT)

    def MutateInt(self, value):
        """Mutates underlying integer value bytes in place.

        Args:
          value: New integer value. It must fit to the byte size of the existing
            encoded value.

        Returns:
          Whether the value was mutated or not.
        """
        if self._type is Type.INT:
            return _Mutate(I, self._buf, value, self._parent_width, BitWidth.I(value))
        elif self._type is Type.INDIRECT_INT:
            return _Mutate(
                I, self._Indirect(), value, self._byte_width, BitWidth.I(value)
            )
        elif self._type is Type.UINT:
            return _Mutate(U, self._buf, value, self._parent_width, BitWidth.U(value))
        elif self._type is Type.INDIRECT_UINT:
            return _Mutate(
                U, self._Indirect(), value, self._byte_width, BitWidth.U(value)
            )
        else:
            return False

    @property
    def IsFloat(self):
        return self._type in (Type.FLOAT, Type.INDIRECT_FLOAT)

    @property
    def AsFloat(self):
        """Returns current reference as floating point value."""
        if self.IsNull:
            return 0.0
        elif self.IsBool:
            return float(self.AsBool)
        elif self.IsInt:
            return float(self.AsInt)
        elif self._type is Type.FLOAT:
            return _Unpack(F, self._Bytes)
        elif self._type is Type.INDIRECT_FLOAT:
            return _Unpack(F, self._Indirect()[: self._byte_width])
        elif self.IsString:
            return float(self.AsString)
        elif self.IsVector:
            return float(len(self.AsVector))
        elif self.IsTypedVector():
            return float(len(self.AsTypedVector))
        elif self.IsFixedTypedVector():
            return float(len(self.FixedTypedVector))
        else:
            raise self._ConvertError(Type.FLOAT)

    def MutateFloat(self, value):
        """Mutates underlying floating point value bytes in place.

        Args:
          value: New float value. It must fit to the byte size of the existing
            encoded value.

        Returns:
          Whether the value was mutated or not.
        """
        if self._type is Type.FLOAT:
            return _Mutate(
                F, self._buf, value, self._parent_width, BitWidth.B(self._parent_width)
            )
        elif self._type is Type.INDIRECT_FLOAT:
            return _Mutate(
                F,
                self._Indirect(),
                value,
                self._byte_width,
                BitWidth.B(self._byte_width),
            )
        else:
            return False

    @property
    def IsKey(self):
        return self._type is Type.KEY

    @property
    def AsKeyBytes(self):
        if self.IsKey:
            return Key(self._Indirect(), self._byte_width).Bytes
        else:
            raise self._ConvertError(Type.KEY)

    @property
    def AsKey(self):
        if self.IsKey:
            return str(Key(self._Indirect(), self._byte_width))
        else:
            raise self._ConvertError(Type.KEY)

    @property
    def IsString(self):
        return self._type is Type.STRING

    @property
    def AsString(self):
        if self.IsString:
            return str(String(self._Indirect(), self._byte_width))
        elif self.IsKey:
            return self.AsKey
        else:
            raise self._ConvertError(Type.STRING)

    def MutateString(self, value):
        return String(self._Indirect(), self._byte_width).Mutate(value)

    @property
    def IsBlob(self):
        return self._type is Type.BLOB

    @property
    def AsBlob(self):
        if self.IsBlob:
            return Blob(self._Indirect(), self._byte_width).Bytes
        else:
            raise self._ConvertError(Type.BLOB)

    @property
    def IsAnyVector(self):
        return self.IsVector or self.IsTypedVector or self.IsFixedTypedVector()

    @property
    def IsVector(self):
        return self._type in (Type.VECTOR, Type.MAP)

    @property
    def AsVector(self):
        if self.IsVector:
            return Vector(self._Indirect(), self._byte_width)
        else:
            raise self._ConvertError(Type.VECTOR)

    @property
    def IsTypedVector(self):
        return Type.IsTypedVector(self._type)

    @property
    def AsTypedVector(self):
        if self.IsTypedVector:
            return TypedVector(
                self._Indirect(),
                self._byte_width,
                Type.ToTypedVectorElementType(self._type),
            )
        else:
            raise self._ConvertError("TYPED_VECTOR")

    @property
    def IsFixedTypedVector(self):
        return Type.IsFixedTypedVector(self._type)

    @property
    def AsFixedTypedVector(self):
        if self.IsFixedTypedVector:
            element_type, size = Type.ToFixedTypedVectorElementType(self._type)
            return TypedVector(self._Indirect(), self._byte_width, element_type, size)
        else:
            raise self._ConvertError("FIXED_TYPED_VECTOR")

    @property
    def IsMap(self):
        return self._type is Type.MAP

    @property
    def AsMap(self):
        if self.IsMap:
            return Map(self._Indirect(), self._byte_width)
        else:
            raise self._ConvertError(Type.MAP)

    @property
    def Value(self):
        """Converts current reference to value of corresponding type.

        This is equivalent to calling `AsInt` for integer values, `AsFloat` for
        floating point values, etc.

        Returns:
          Value of corresponding type.
        """
        if self.IsNull:
            return None
        elif self.IsBool:
            return self.AsBool
        elif self.IsInt:
            return self.AsInt
        elif self.IsFloat:
            return self.AsFloat
        elif self.IsString:
            return self.AsString
        elif self.IsKey:
            return self.AsKey
        elif self.IsBlob:
            return self.AsBlob
        elif self.IsMap:
            return self.AsMap.Value
        elif self.IsVector:
            return self.AsVector.Value
        elif self.IsTypedVector:
            return self.AsTypedVector.Value
        elif self.IsFixedTypedVector:
            return self.AsFixedTypedVector.Value
        else:
            raise TypeError("cannot convert %r to value" % self)


def _IsIterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


class Value:
    """Class to represent given value during the encoding process."""

    @staticmethod
    def Null():
        return Value(0, Type.NULL, BitWidth.W8)

    @staticmethod
    def Bool(value):
        return Value(value, Type.BOOL, BitWidth.W8)

    @staticmethod
    def Int(value, bit_width):
        return Value(value, Type.INT, bit_width)

    @staticmethod
    def UInt(value, bit_width):
        return Value(value, Type.UINT, bit_width)

    @staticmethod
    def Float(value, bit_width):
        return Value(value, Type.FLOAT, bit_width)

    @staticmethod
    def Key(offset):
        return Value(offset, Type.KEY, BitWidth.W8)

    def __init__(self, value, type_, min_bit_width):
        self._value = value
        self._type = type_

        # For scalars: of itself, for vector: of its elements, for string: length.
        self._min_bit_width = min_bit_width

    @property
    def Value(self):
        return self._value

    @property
    def Type(self):
        return self._type

    @property
    def MinBitWidth(self):
        return self._min_bit_width

    def StoredPackedType(self, parent_bit_width=BitWidth.W8):
        return Type.Pack(self._type, self.StoredWidth(parent_bit_width))

    # We have an absolute offset, but want to store a relative offset
    # elem_index elements beyond the current buffer end. Since whether
    # the relative offset fits in a certain byte_width depends on
    # the size of the elements before it (and their alignment), we have
    # to test for each size in turn.
    def ElemWidth(self, buf_size, elem_index=0):
        if Type.IsInline(self._type):
            return self._min_bit_width
        for byte_width in 1, 2, 4, 8:
            offset_loc = (
                buf_size + _PaddingBytes(buf_size, byte_width) + elem_index * byte_width
            )
            bit_width = BitWidth.U(offset_loc - self._value)
            if byte_width == (1 << bit_width):
                return bit_width
        raise ValueError("relative offset is too big")

    def StoredWidth(self, parent_bit_width=BitWidth.W8):
        if Type.IsInline(self._type):
            return max(self._min_bit_width, parent_bit_width)
        return self._min_bit_width

    def __repr__(self):
        return "Value(%s, %s, %s)" % (self._value, self._type, self._min_bit_width)

    def __str__(self):
        return str(self._value)


def InMap(func):
    def wrapper(self, *args, **kwargs):
        if isinstance(args[0], str):
            self.Key(args[0])
            func(self, *args[1:], **kwargs)
        else:
            func(self, *args, **kwargs)

    return wrapper


def InMapForString(func):
    def wrapper(self, *args):
        if len(args) == 1:
            func(self, args[0])
        elif len(args) == 2:
            self.Key(args[0])
            func(self, args[1])
        else:
            raise ValueError("invalid number of arguments")

    return wrapper


class Pool:
    """Collection of (data, offset) pairs sorted by data for quick access."""

    def __init__(self):
        self._pool = []  # sorted list of (data, offset) tuples

    def FindOrInsert(self, data, offset):
        do = data, offset
        index = _BinarySearch(self._pool, do, lambda a, b: a[0] < b[0])
        if index != -1:
            _, offset = self._pool[index]
            return offset
        self._pool.insert(index, do)
        return None

    def Clear(self):
        self._pool = []

    @property
    def Elements(self):
        return [data for data, _ in self._pool]


class Builder:
    """Helper class to encode structural data into flexbuffers format."""

    def __init__(
        self, share_strings=False, share_keys=True, force_min_bit_width=BitWidth.W8
    ):
        self._share_strings = share_strings
        self._share_keys = share_keys
        self._force_min_bit_width = force_min_bit_width

        self._string_pool = Pool()
        self._key_pool = Pool()

        self._finished = False
        self._buf = bytearray()
        self._stack = []

    def __len__(self):
        return len(self._buf)

    @property
    def StringPool(self):
        return self._string_pool

    @property
    def KeyPool(self):
        return self._key_pool

    def Clear(self):
        self._string_pool.Clear()
        self._key_pool.Clear()
        self._finished = False
        self._buf = bytearray()
        self._stack = []

    def Finish(self):
        """Finishes encoding process and returns underlying buffer."""
        if self._finished:
            raise RuntimeError("builder has been already finished")

        # If you hit this exception, you likely have objects that were never
        # included in a parent. You need to have exactly one root to finish a
        # buffer. Check your Start/End calls are matched, and all objects are inside
        # some other object.
        if len(self._stack) != 1:
            raise RuntimeError("internal stack size must be one")

        value = self._stack[0]
        byte_width = self._Align(value.ElemWidth(len(self._buf)))
        self._WriteAny(value, byte_width=byte_width)  # Root value
        self._Write(U, value.StoredPackedType(), byte_width=1)  # Root type
        self._Write(U, byte_width, byte_width=1)  # Root size

        self.finished = True
        return self._buf

    def _ReadKey(self, offset):
        key = self._buf[offset:]
        return key[: key.find(0)]

    def _Align(self, alignment):
        byte_width = 1 << alignment
        self._buf.extend(b"\x00" * _PaddingBytes(len(self._buf), byte_width))
        return byte_width

    def _Write(self, fmt, value, byte_width):
        self._buf.extend(_Pack(fmt, value, byte_width))

    def _WriteVector(self, fmt, values, byte_width):
        self._buf.extend(_PackVector(fmt, values, byte_width))

    def _WriteOffset(self, offset, byte_width):
        relative_offset = len(self._buf) - offset
        assert byte_width == 8 or relative_offset < (1 << (8 * byte_width))
        self._Write(U, relative_offset, byte_width)

    def _WriteAny(self, value, byte_width):
        fmt = {
            Type.NULL: U,
            Type.BOOL: U,
            Type.INT: I,
            Type.UINT: U,
            Type.FLOAT: F,
        }.get(value.Type)
        if fmt:
            self._Write(fmt, value.Value, byte_width)
        else:
            self._WriteOffset(value.Value, byte_width)

    def _WriteBlob(self, data, append_zero, type_):
        bit_width = BitWidth.U(len(data))
        byte_width = self._Align(bit_width)
        self._Write(U, len(data), byte_width)
        loc = len(self._buf)
        self._buf.extend(data)
        if append_zero:
            self._buf.append(0)
        self._stack.append(Value(loc, type_, bit_width))
        return loc

    def _WriteScalarVector(self, element_type, byte_width, elements, fixed):
        """Writes scalar vector elements to the underlying buffer."""
        bit_width = BitWidth.B(byte_width)
        # If you get this exception, you're trying to write a vector with a size
        # field that is bigger than the scalars you're trying to write (e.g. a
        # byte vector > 255 elements). For such types, write a "blob" instead.
        if BitWidth.U(len(elements)) > bit_width:
            raise ValueError("too many elements for the given byte_width")

        self._Align(bit_width)
        if not fixed:
            self._Write(U, len(elements), byte_width)

        loc = len(self._buf)

        fmt = {Type.INT: I, Type.UINT: U, Type.FLOAT: F}.get(element_type)
        if not fmt:
            raise TypeError("unsupported element_type")
        self._WriteVector(fmt, elements, byte_width)

        type_ = Type.ToTypedVector(element_type, len(elements) if fixed else 0)
        self._stack.append(Value(loc, type_, bit_width))
        return loc

    def _CreateVector(self, elements, typed, fixed, keys=None):
        """Writes vector elements to the underlying buffer."""
        length = len(elements)

        if fixed and not typed:
            raise ValueError("fixed vector must be typed")

        # Figure out smallest bit width we can store this vector with.
        bit_width = max(self._force_min_bit_width, BitWidth.U(length))
        prefix_elems = 1  # Vector size
        if keys:
            bit_width = max(bit_width, keys.ElemWidth(len(self._buf)))
            prefix_elems += 2  # Offset to the keys vector and its byte width.

        vector_type = Type.KEY
        # Check bit widths and types for all elements.
        for i, e in enumerate(elements):
            bit_width = max(bit_width, e.ElemWidth(len(self._buf), prefix_elems + i))

            if typed:
                if i == 0:
                    vector_type = e.Type
                else:
                    if vector_type != e.Type:
                        raise RuntimeError(
                            "typed vector elements must be of the same type"
                        )

        if fixed and not Type.IsFixedTypedVectorElementType(vector_type):
            raise RuntimeError("must be fixed typed vector element type")

        byte_width = self._Align(bit_width)
        # Write vector. First the keys width/offset if available, and size.
        if keys:
            self._WriteOffset(keys.Value, byte_width)
            self._Write(U, 1 << keys.MinBitWidth, byte_width)

        if not fixed:
            self._Write(U, length, byte_width)

        # Then the actual data.
        loc = len(self._buf)
        for e in elements:
            self._WriteAny(e, byte_width)

        # Then the types.
        if not typed:
            for e in elements:
                self._buf.append(e.StoredPackedType(bit_width))

        if keys:
            type_ = Type.MAP
        else:
            if typed:
                type_ = Type.ToTypedVector(vector_type, length if fixed else 0)
            else:
                type_ = Type.VECTOR

        return Value(loc, type_, bit_width)

    def _PushIndirect(self, value, type_, bit_width):
        byte_width = self._Align(bit_width)
        loc = len(self._buf)
        fmt = {Type.INDIRECT_INT: I, Type.INDIRECT_UINT: U, Type.INDIRECT_FLOAT: F}[
            type_
        ]
        self._Write(fmt, value, byte_width)
        self._stack.append(Value(loc, type_, bit_width))

    @InMapForString
    def String(self, value):
        """Encodes string value."""
        reset_to = len(self._buf)
        encoded = value.encode("utf-8")
        loc = self._WriteBlob(encoded, append_zero=True, type_=Type.STRING)
        if self._share_strings:
            prev_loc = self._string_pool.FindOrInsert(encoded, loc)
            if prev_loc is not None:
                del self._buf[reset_to:]
                self._stack[-1]._value = loc = prev_loc  # pylint: disable=protected-access

        return loc

    @InMap
    def Blob(self, value):
        """Encodes binary blob value.

        Args:
          value: A byte/bytearray value to encode

        Returns:
          Offset of the encoded value in underlying the byte buffer.
        """
        return self._WriteBlob(value, append_zero=False, type_=Type.BLOB)

    def Key(self, value):
        """Encodes key value.

        Args:
          value: A byte/bytearray/str value to encode. Byte object must not contain
            zero bytes. String object must be convertible to ASCII.

        Returns:
          Offset of the encoded value in the underlying byte buffer.
        """
        if isinstance(value, (bytes, bytearray)):
            encoded = value
        else:
            encoded = value.encode("ascii")

        if 0 in encoded:
            raise ValueError("key contains zero byte")

        loc = len(self._buf)
        self._buf.extend(encoded)
        self._buf.append(0)
        if self._share_keys:
            prev_loc = self._key_pool.FindOrInsert(encoded, loc)
            if prev_loc is not None:
                del self._buf[loc:]
                loc = prev_loc

        self._stack.append(Value.Key(loc))
        return loc

    def Null(self, key=None):
        """Encodes None value."""
        if key:
            self.Key(key)
        self._stack.append(Value.Null())

    @InMap
    def Bool(self, value):
        """Encodes boolean value.

        Args:
          value: A boolean value.
        """
        self._stack.append(Value.Bool(value))

    @InMap
    def Int(self, value, byte_width=0):
        """Encodes signed integer value.

        Args:
          value: A signed integer value.
          byte_width: Number of bytes to use: 1, 2, 4, or 8.
        """
        bit_width = BitWidth.I(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._stack.append(Value.Int(value, bit_width))

    @InMap
    def IndirectInt(self, value, byte_width=0):
        """Encodes signed integer value indirectly.

        Args:
          value: A signed integer value.
          byte_width: Number of bytes to use: 1, 2, 4, or 8.
        """
        bit_width = BitWidth.I(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._PushIndirect(value, Type.INDIRECT_INT, bit_width)

    @InMap
    def UInt(self, value, byte_width=0):
        """Encodes unsigned integer value.

        Args:
          value: An unsigned integer value.
          byte_width: Number of bytes to use: 1, 2, 4, or 8.
        """
        bit_width = BitWidth.U(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._stack.append(Value.UInt(value, bit_width))

    @InMap
    def IndirectUInt(self, value, byte_width=0):
        """Encodes unsigned integer value indirectly.

        Args:
          value: An unsigned integer value.
          byte_width: Number of bytes to use: 1, 2, 4, or 8.
        """
        bit_width = BitWidth.U(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._PushIndirect(value, Type.INDIRECT_UINT, bit_width)

    @InMap
    def Float(self, value, byte_width=0):
        """Encodes floating point value.

        Args:
          value: A floating point value.
          byte_width: Number of bytes to use: 4 or 8.
        """
        bit_width = BitWidth.F(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._stack.append(Value.Float(value, bit_width))

    @InMap
    def IndirectFloat(self, value, byte_width=0):
        """Encodes floating point value indirectly.

        Args:
          value: A floating point value.
          byte_width: Number of bytes to use: 4 or 8.
        """
        bit_width = BitWidth.F(value) if byte_width == 0 else BitWidth.B(byte_width)
        self._PushIndirect(value, Type.INDIRECT_FLOAT, bit_width)

    def _StartVector(self):
        """Starts vector construction."""
        return len(self._stack)

    def _EndVector(self, start, typed, fixed):
        """Finishes vector construction by encodung its elements."""
        vec = self._CreateVector(self._stack[start:], typed, fixed)
        del self._stack[start:]
        self._stack.append(vec)
        return vec.Value

    @contextlib.contextmanager
    def Vector(self, key=None):
        if key:
            self.Key(key)

        try:
            start = self._StartVector()
            yield self
        finally:
            self._EndVector(start, typed=False, fixed=False)

    @InMap
    def VectorFromElements(self, elements):
        """Encodes sequence of any elements as a vector.

        Args:
          elements: sequence of elements, they may have different types.
        """
        with self.Vector():
            for e in elements:
                self.Add(e)

    @contextlib.contextmanager
    def TypedVector(self, key=None):
        if key:
            self.Key(key)

        try:
            start = self._StartVector()
            yield self
        finally:
            self._EndVector(start, typed=True, fixed=False)

    @InMap
    def TypedVectorFromElements(self, elements, element_type=None):
        """Encodes sequence of elements of the same type as typed vector.

        Args:
          elements: Sequence of elements, they must be of the same type.
          element_type: Suggested element type. Setting it to None means determining
            correct value automatically based on the given elements.
        """
        if isinstance(elements, array.array):
            if elements.typecode == "f":
                self._WriteScalarVector(Type.FLOAT, 4, elements, fixed=False)
            elif elements.typecode == "d":
                self._WriteScalarVector(Type.FLOAT, 8, elements, fixed=False)
            elif elements.typecode in ("b", "h", "i", "l", "q"):
                self._WriteScalarVector(
                    Type.INT, elements.itemsize, elements, fixed=False
                )
            elif elements.typecode in ("B", "H", "I", "L", "Q"):
                self._WriteScalarVector(
                    Type.UINT, elements.itemsize, elements, fixed=False
                )
            else:
                raise ValueError("unsupported array typecode: %s" % elements.typecode)
        else:
            add = self.Add if element_type is None else self.Adder(element_type)
            with self.TypedVector():
                for e in elements:
                    add(e)

    @InMap
    def FixedTypedVectorFromElements(self, elements, element_type=None, byte_width=0):
        """Encodes sequence of elements of the same type as fixed typed vector.

        Args:
          elements: Sequence of elements, they must be of the same type. Allowed
            types are `Type.INT`, `Type.UINT`, `Type.FLOAT`. Allowed number of
            elements are 2, 3, or 4.
          element_type: Suggested element type. Setting it to None means determining
            correct value automatically based on the given elements.
          byte_width: Number of bytes to use per element. For `Type.INT` and
            `Type.UINT`: 1, 2, 4, or 8. For `Type.FLOAT`: 4 or 8. Setting it to 0
            means determining correct value automatically based on the given
            elements.
        """
        if not 2 <= len(elements) <= 4:
            raise ValueError("only 2, 3, or 4 elements are supported")

        types = {type(e) for e in elements}
        if len(types) != 1:
            raise TypeError("all elements must be of the same type")

        (type_,) = types

        if element_type is None:
            element_type = {int: Type.INT, float: Type.FLOAT}.get(type_)
            if not element_type:
                raise TypeError("unsupported element_type: %s" % type_)

        if byte_width == 0:
            width = {
                Type.UINT: BitWidth.U,
                Type.INT: BitWidth.I,
                Type.FLOAT: BitWidth.F,
            }[element_type]
            byte_width = 1 << max(width(e) for e in elements)

        self._WriteScalarVector(element_type, byte_width, elements, fixed=True)

    def _StartMap(self):
        """Starts map construction."""
        return len(self._stack)

    def _EndMap(self, start):
        """Finishes map construction by encodung its elements."""
        # Interleaved keys and values on the stack.
        stack = self._stack[start:]

        if len(stack) % 2 != 0:
            raise RuntimeError("must be even number of keys and values")

        for key in stack[::2]:
            if key.Type is not Type.KEY:
                raise RuntimeError("all map keys must be of %s type" % Type.KEY)

        pairs = zip(stack[::2], stack[1::2])  # [(key, value), ...]
        pairs = sorted(pairs, key=lambda pair: self._ReadKey(pair[0].Value))

        del self._stack[start:]
        for pair in pairs:
            self._stack.extend(pair)

        keys = self._CreateVector(self._stack[start::2], typed=True, fixed=False)
        values = self._CreateVector(
            self._stack[start + 1 :: 2], typed=False, fixed=False, keys=keys
        )

        del self._stack[start:]
        self._stack.append(values)
        return values.Value

    @contextlib.contextmanager
    def Map(self, key=None):
        if key:
            self.Key(key)

        try:
            start = self._StartMap()
            yield self
        finally:
            self._EndMap(start)

    def MapFromElements(self, elements):
        start = self._StartMap()
        for k, v in elements.items():
            self.Key(k)
            self.Add(v)
        self._EndMap(start)

    def Adder(self, type_):
        return {
            Type.BOOL: self.Bool,
            Type.INT: self.Int,
            Type.INDIRECT_INT: self.IndirectInt,
            Type.UINT: self.UInt,
            Type.INDIRECT_UINT: self.IndirectUInt,
            Type.FLOAT: self.Float,
            Type.INDIRECT_FLOAT: self.IndirectFloat,
            Type.KEY: self.Key,
            Type.BLOB: self.Blob,
            Type.STRING: self.String,
        }[type_]

    @InMapForString
    def Add(self, value):
        """Encodes value of any supported type."""
        if value is None:
            self.Null()
        elif isinstance(value, bool):
            self.Bool(value)
        elif isinstance(value, int):
            self.Int(value)
        elif isinstance(value, float):
            self.Float(value)
        elif isinstance(value, str):
            self.String(value)
        elif isinstance(value, (bytes, bytearray)):
            self.Blob(value)
        elif isinstance(value, dict):
            with self.Map():
                for k, v in value.items():
                    self.Key(k)
                    self.Add(v)
        elif isinstance(value, array.array):
            self.TypedVectorFromElements(value)
        elif _IsIterable(value):
            self.VectorFromElements(value)
        else:
            raise TypeError("unsupported python type: %s" % type(value))

    @property
    def LastValue(self):
        return self._stack[-1]

    @InMap
    def ReuseValue(self, value):
        self._stack.append(value)


def GetRoot(buf):
    """Returns root `Ref` object for the given buffer."""
    if len(buf) < 3:
        raise ValueError("buffer is too small")
    byte_width = buf[-1]
    return Ref.PackedType(Buf(buf, -(2 + byte_width)), byte_width, packed_type=buf[-2])


def Dumps(obj):
    """Returns bytearray with the encoded python object."""
    fbb = Builder()
    fbb.Add(obj)
    return fbb.Finish()


def Loads(buf):
    """Returns python object decoded from the buffer."""
    return GetRoot(buf).Value



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: models

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class HelloReply(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = HelloReply()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsHelloReply(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # HelloReply
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # HelloReply
    def Message(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def HelloReplyStart(builder):
    builder.StartObject(1)

def Start(builder):
    HelloReplyStart(builder)

def HelloReplyAddMessage(builder, message):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(message), 0)

def AddMessage(builder, message):
    HelloReplyAddMessage(builder, message)

def HelloReplyEnd(builder):
    return builder.EndObject()

def End(builder):
    return HelloReplyEnd(builder)



import argparse
import sys
import grpc

sys.path.insert(0, "../../../../../flatbuffers/python")

import flatbuffers
from models import HelloReply, HelloRequest, greeter_grpc_fb

parser = argparse.ArgumentParser()
parser.add_argument("port", help="server port to connect to", default=3000)
parser.add_argument(
    "name", help="name to be sent to server", default="flatbuffers"
)


def say_hello(stub, hello_request):
  reply = stub.SayHello(hello_request)
  r = HelloReply.HelloReply.GetRootAs(reply)
  print(r.Message())


def say_many_hellos(stub, hello_request):
  greetings = stub.SayManyHellos(hello_request)
  for greeting in greetings:
    r = HelloReply.HelloReply.GetRootAs(greeting)
    print(r.Message())


def main():
  args = parser.parse_args()

  with grpc.insecure_channel("localhost:" + args.port) as channel:
    builder = flatbuffers.Builder()
    ind = builder.CreateString(args.name)
    HelloRequest.HelloRequestStart(builder)
    HelloRequest.HelloRequestAddName(builder, ind)
    root = HelloRequest.HelloRequestEnd(builder)
    builder.Finish(root)
    output = bytes(builder.Output())
    stub = greeter_grpc_fb.GreeterStub(channel)
    say_hello(stub, output)
    say_many_hellos(stub, output)


main()



#!/usr/bin/env python3
#
# Copyright 2021 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import platform
import subprocess
import sys

# Get the path where this script is located so we can invoke the script from
# any directory and have the paths work correctly.
script_path = Path(__file__).parent.resolve()

# Get the root path as an absolute path, so all derived paths are absolute.
root_path = script_path.parent.absolute()

result = subprocess.run(
    ["git", "diff", "--quiet", "--ignore-cr-at-eol"], cwd=root_path
)

if result.returncode != 0:
  print(
      "\n"
      "ERROR: *********************************************************\n"
      "ERROR: * The following differences were found after building.  *\n"
      "ERROR: * Perhaps there is a difference in the flags for the.   *\n"
      "ERROR: * CMakeLists.txt vs the script/generate_code.py script? *\n"
      "ERROR: *********************************************************\n"
  )
  subprocess.run(["git", "diff", "--binary", "--exit-code"], cwd=root_path)
  sys.exit(result.returncode)

# Rung the generate_code.py script, forwarding arguments
gen_cmd = ["scripts/generate_code.py"] + sys.argv[1:]
if platform.system() == "Windows":
  gen_cmd = ["py"] + gen_cmd
subprocess.run(gen_cmd, cwd=root_path)

result = subprocess.run(
    ["git", "diff", "--quiet", "--ignore-cr-at-eol"], cwd=root_path
)

if result.returncode != 0:
  print(
      "\n"
      "ERROR: ********************************************************\n"
      "ERROR: * The following differences were found after running   *\n"
      "ERROR: * the script/generate_code.py script. Maybe you forgot *\n"
      "ERROR: * to run it after making changes in a generator?       *\n"
      "ERROR: ********************************************************\n"
  )
  subprocess.run(["git", "diff", "--binary", "--exit-code"], cwd=root_path)
  sys.exit(result.returncode)

sys.exit(0)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A tiny version of `six` to help with backwards compability.

Also includes compatibility helpers for numpy.
"""

import sys

PY2 = sys.version_info[0] == 2
PY26 = sys.version_info[0:2] == (2, 6)
PY27 = sys.version_info[0:2] == (2, 7)
PY275 = sys.version_info[0:3] >= (2, 7, 5)
PY3 = sys.version_info[0] == 3
PY34 = sys.version_info[0:2] >= (3, 4)

if PY3:
  import importlib.machinery

  string_types = (str,)
  binary_types = (bytes, bytearray)
  range_func = range
  memoryview_type = memoryview
  struct_bool_decl = "?"
else:
  import imp

  string_types = (unicode,)
  if PY26 or PY27:
    binary_types = (str, bytearray)
  else:
    binary_types = (str,)
  range_func = xrange
  if PY26 or (PY27 and not PY275):
    memoryview_type = buffer
    struct_bool_decl = "<b"
  else:
    memoryview_type = memoryview
    struct_bool_decl = "?"

# Helper functions to facilitate making numpy optional instead of required


def import_numpy():
  """Returns the numpy module if it exists on the system,

  otherwise returns None.
  """
  if PY3:
    numpy_exists = importlib.machinery.PathFinder.find_spec("numpy") is not None
  else:
    try:
      imp.find_module("numpy")
      numpy_exists = True
    except ImportError:
      numpy_exists = False

  if numpy_exists:
    # We do this outside of try/except block in case numpy exists
    # but is not installed correctly. We do not want to catch an
    # incorrect installation which would manifest as an
    # ImportError.
    import numpy as np
  else:
    np = None

  return np


class NumpyRequiredForThisFeature(RuntimeError):
  """Error raised when user tries to use a feature that

  requires numpy without having numpy installed.
  """

  pass


# NOTE: Future Jython support may require code here (look at `six`).



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from conans import CMake, ConanFile


class TestPackageConan(ConanFile):
  settings = "os", "compiler", "build_type", "arch"
  generators = "cmake"

  def build(self):
    cmake = CMake(self)
    cmake.configure()
    cmake.build()

  def test(self):
    bin_path = os.path.join("bin", "test_package")
    self.run(bin_path, run_environment=True)
    self.run("flatc --version", run_environment=True)
    self.run("flathash fnv1_16 conan", run_environment=True)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Field(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Field()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsField(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def FieldBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x42\x46\x42\x53", size_prefixed=size_prefixed)

    # Field
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Field
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Field
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from reflection.Type import Type
            obj = Type()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Field
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Field
    def Offset(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Field
    def DefaultInteger(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # Field
    def DefaultReal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Field
    def Deprecated(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Field
    def Required(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Field
    def Key(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Field
    def Attributes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.KeyValue import KeyValue
            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Field
    def AttributesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Field
    def AttributesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # Field
    def Documentation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Field
    def DocumentationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Field
    def DocumentationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # Field
    def Optional(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Number of padding octets to always add after this field. Structs only.
    # Field
    def Padding(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # If the field uses 64-bit offsets.
    # Field
    def Offset64(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def FieldStart(builder):
    builder.StartObject(14)

def Start(builder):
    FieldStart(builder)

def FieldAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    FieldAddName(builder, name)

def FieldAddType(builder, type):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(type), 0)

def AddType(builder, type):
    FieldAddType(builder, type)

def FieldAddId(builder, id):
    builder.PrependUint16Slot(2, id, 0)

def AddId(builder, id):
    FieldAddId(builder, id)

def FieldAddOffset(builder, offset):
    builder.PrependUint16Slot(3, offset, 0)

def AddOffset(builder, offset):
    FieldAddOffset(builder, offset)

def FieldAddDefaultInteger(builder, defaultInteger):
    builder.PrependInt64Slot(4, defaultInteger, 0)

def AddDefaultInteger(builder, defaultInteger):
    FieldAddDefaultInteger(builder, defaultInteger)

def FieldAddDefaultReal(builder, defaultReal):
    builder.PrependFloat64Slot(5, defaultReal, 0.0)

def AddDefaultReal(builder, defaultReal):
    FieldAddDefaultReal(builder, defaultReal)

def FieldAddDeprecated(builder, deprecated):
    builder.PrependBoolSlot(6, deprecated, 0)

def AddDeprecated(builder, deprecated):
    FieldAddDeprecated(builder, deprecated)

def FieldAddRequired(builder, required):
    builder.PrependBoolSlot(7, required, 0)

def AddRequired(builder, required):
    FieldAddRequired(builder, required)

def FieldAddKey(builder, key):
    builder.PrependBoolSlot(8, key, 0)

def AddKey(builder, key):
    FieldAddKey(builder, key)

def FieldAddAttributes(builder, attributes):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(attributes), 0)

def AddAttributes(builder, attributes):
    FieldAddAttributes(builder, attributes)

def FieldStartAttributesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartAttributesVector(builder, numElems):
    return FieldStartAttributesVector(builder, numElems)

def FieldAddDocumentation(builder, documentation):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(documentation), 0)

def AddDocumentation(builder, documentation):
    FieldAddDocumentation(builder, documentation)

def FieldStartDocumentationVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartDocumentationVector(builder, numElems):
    return FieldStartDocumentationVector(builder, numElems)

def FieldAddOptional(builder, optional):
    builder.PrependBoolSlot(11, optional, 0)

def AddOptional(builder, optional):
    FieldAddOptional(builder, optional)

def FieldAddPadding(builder, padding):
    builder.PrependUint16Slot(12, padding, 0)

def AddPadding(builder, padding):
    FieldAddPadding(builder, padding)

def FieldAddOffset64(builder, offset64):
    builder.PrependBoolSlot(13, offset64, 0)

def AddOffset64(builder, offset64):
    FieldAddOffset64(builder, offset64)

def FieldEnd(builder):
    return builder.EndObject()

def End(builder):
    return FieldEnd(builder)



#!/usr/bin/env python3
# Copyright (c) typedef int GmbH, Germany, 2025. All rights reserved.
#
# Smoke tests for zlmdb package verification.
# Used by CI to verify wheels and sdists actually work after building.

"""
Smoke tests for zlmdb package.

This script verifies that a zlmdb installation is functional by testing:
1. Import zlmdb and check version
2. Import zlmdb.lmdb (CFFI extension) and check version
3. Basic LMDB operations (open/put/get/close)
4. Import zlmdb.flatbuffers and check version (shared test)
5. Verify flatc binary is available and executable (shared test)
6. Verify reflection files are present (shared test)

ALL TESTS ARE REQUIRED. Both wheel installs and sdist installs MUST
provide identical functionality including the flatc binary and
reflection.bfbs file.

Note: FlatBuffers/flatc tests use shared functions from smoke_test_flatc.py
      (source: wamp-cicd/scripts/flatc/smoke_test_flatc.py)
"""

import sys
import tempfile

# Import shared FlatBuffers test functions
from smoke_test_flatc import (
    test_import_flatbuffers,
    test_flatc_binary,
    test_reflection_files,
)

# Package name for shared tests
PACKAGE_NAME = "zlmdb"


def test_import_zlmdb():
    """Test 1: Import zlmdb and check version."""
    print("Test 1: Importing zlmdb and checking version...")
    try:
        import zlmdb
        print(f"  zlmdb version: {zlmdb.__version__}")
        print("  PASS")
        return True
    except Exception as e:
        print(f"  FAIL: Could not import zlmdb: {e}")
        return False


def test_import_lmdb():
    """Test 2: Import zlmdb.lmdb (CFFI extension) and check version."""
    print("Test 2: Importing zlmdb.lmdb (CFFI extension)...")
    try:
        import zlmdb.lmdb as lmdb
        print(f"  LMDB version: {lmdb.version()}")
        print("  PASS")
        return True
    except Exception as e:
        print(f"  FAIL: Could not import zlmdb.lmdb: {e}")
        return False


def test_lmdb_operations():
    """Test 3: Basic LMDB operations (open/put/get/close)."""
    print("Test 3: Basic LMDB operations (open/put/get/close)...")
    try:
        import zlmdb.lmdb as lmdb

        with tempfile.TemporaryDirectory() as tmpdir:
            env = lmdb.open(tmpdir, max_dbs=1)
            with env.begin(write=True) as txn:
                txn.put(b"key1", b"value1")
                txn.put(b"key2", b"value2")
            with env.begin() as txn:
                assert txn.get(b"key1") == b"value1", "key1 mismatch"
                assert txn.get(b"key2") == b"value2", "key2 mismatch"
                assert txn.get(b"key3") is None, "key3 should be None"
            env.close()
        print("  LMDB operations: open, put, get, close - all work!")
        print("  PASS")
        return True
    except Exception as e:
        print(f"  FAIL: LMDB operations failed: {e}")
        return False


def main():
    """Run all smoke tests."""
    print("=" * 72)
    print("  SMOKE TESTS - Verifying zlmdb installation")
    print("=" * 72)
    print()
    print(f"Python: {sys.version}")
    print()

    # All tests are REQUIRED - sdist MUST provide same functionality as wheels
    # Tests 4-6 use shared functions from smoke_test_flatc.py
    tests = [
        ("Test 1", test_import_zlmdb),
        ("Test 2", test_import_lmdb),
        ("Test 3", test_lmdb_operations),
        ("Test 4", lambda: test_import_flatbuffers(PACKAGE_NAME)),
        ("Test 5", lambda: test_flatc_binary(PACKAGE_NAME)),
        ("Test 6", lambda: test_reflection_files(PACKAGE_NAME)),
    ]

    failures = 0
    passed = 0

    for name, test in tests:
        result = test()
        if result is True:
            passed += 1
        else:
            failures += 1
        print()

    total = len(tests)
    print("=" * 72)
    if failures == 0:
        print(f"ALL SMOKE TESTS PASSED ({passed}/{total})")
        print("=" * 72)
        return 0
    else:
        print(f"SMOKE TESTS FAILED ({passed} passed, {failures} failed)")
        print("=" * 72)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import datetime
import pickle
import json
import cbor2
from typing import Optional, List, Dict

RESULT = {
    'objects': 0,
    'bytes': 0
}


class Tag(object):
    GEEK = 1
    VIP = 2

# requires python 3.6+:
#
# class User(object):
#     oid: int
#     name: str
#     authid: str
#     email: str
#     birthday: datetime.date
#     is_friendly: bool
#     tags: Optional[List[str]]
#     ratings: Dict[str, float] = {}
#     friends: List[int] = []
#     referred_by: int = None


class User(object):
    def __init__(self):
        self.oid = None
        self.name = None
        self.authid = None
        self.email = None
        self.birthday = None
        self.is_friendly = None
        self.tags = None
        self.ratings = {}
        self.friends = []
        self.referred_by = None

    def marshal(self):
        obj = {
            'oid': self.oid,
            'name': self.name,
            'authid': self.authid,
            'email': self.email,
            'birthday': self.birthday,
            'is_friendly': self.is_friendly,
            'tags': self.tags,
            'ratings': self.ratings,
            'friends': self.friends,
            'referred_by': self.referred_by,
        }
        return obj

    @staticmethod
    def parse(obj):
        user = User()
        user.oid = obj.get('oid', None)
        user.name = obj.get('name', None)
        user.authid = obj.get('authid', None)
        user.email = obj.get('email', None)
        user.birthday = obj.get('birthday', None)
        user.is_friendly = obj.get('is_friendly', None)
        user.tags = obj.get('tags', None)
        user.ratings = obj.get('ratings', {})
        user.friends = obj.get('friends', [])
        user.referred_by = obj.get('referred_by', None)
        return user


def test():
    global RESULT

    user = User()
    user.oid = 23
    user.name = 'Homer Simpson'
    user.authid = 'homer'
    user.email = 'homer.simpson@example.com'
    user.birthday = {
        'year': 1950,
        'month': 12,
        'day': 24
    }
    user.is_friendly = True
    user.tags = [Tag.GEEK, Tag.VIP]

    #data = json.dumps(user.marshal(), ensure_ascii=False)
    data = cbor2.dumps(user.marshal())

    RESULT['objects'] += 1
    RESULT['bytes'] += len(data)


import timeit

N = 1000
M = 100000

for i in range(N):
    secs = timeit.timeit(test, number=M)
    ops = round(float(M) / secs, 1)
    print('{} objects/sec'.format(ops))
    print(RESULT)



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from . import util
from ._git_version import __git_version__
from ._version import __version__
from .builder import Builder
from .compat import range_func as compat_range
from .table import Table


def version() -> tuple[int, int, int, int | None, str | None]:
    """
    Return the exact git version of the vendored FlatBuffers runtime.

    Handles:

    1. "v25.9.23"              -> (25, 9, 23, None, None)       # Release (Named Tag, CalVer Year.Month.Day)
    2. "v25.9.23-71"           -> (25, 9, 23, 71, None)         # 71 commits ahead of the Release v25.9.23
    3. "v25.9.23-71-g19b2300f" -> (25, 9, 23, 71, "19b2300f")   # dito, with Git commit hash
    """

    # Pattern explanation:
    # ^v                : Start of string, literal 'v'
    # (\d+)\.(\d+)\.(\d+) : Groups 1,2,3 - Major.Minor.Patch (Required)
    #
    # (?: ... )?        : Non-capturing group (grouping only, not saved), optional '?'
    # -(\d+)            : Literal hyphen, Group 4 (Commits)
    #
    # (?: ... )?        : Non-capturing group, optional '?'
    # -g                : Literal hyphen and 'g' (separator)
    # ([0-9a-f]+)       : Group 5 (Hash)

    pattern = r"^v(\d+)\.(\d+)\.(\d+)(?:-(\d+))?(?:-g([0-9a-f]+))?$"

    match = re.match(pattern, __git_version__)

    if match:
        major, minor, patch, commits, commit_hash = match.groups()

        # Convert commits to int if present, else None
        commits_int = int(commits) if commits else None

        return (int(major), int(minor), int(patch), commits_int, commit_hash)

    # Fallback if regex fails entirely (returns 0.0.0 to satisfy type hint)
    return (0, 0, 0, None, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import subprocess
from cpt.packager import ConanMultiPackager


def get_branch():
  try:
    for line in (
        subprocess.check_output("git branch", shell=True).decode().splitlines()
    ):
      line = line.strip()
      if line.startswith("*") and " (HEAD detached" not in line:
        return line.replace("*", "", 1).strip()
    return ""
  except Exception:
    pass
  return ""


def get_version():
  version = get_branch()
  match = re.search(r"v(\d+\.\d+\.\d+.*)", version)
  if match:
    return match.group(1)
  return version


def get_reference(username):
  return "flatbuffers/{}@google/stable".format(get_version())


if __name__ == "__main__":
  login_username = os.getenv("CONAN_LOGIN_USERNAME", "aardappel")
  username = os.getenv("CONAN_USERNAME", "google")
  upload = os.getenv(
      "CONAN_UPLOAD", "https://api.bintray.com/conan/aardappel/flatbuffers"
  )
  stable_branch_pattern = os.getenv(
      "CONAN_STABLE_BRANCH_PATTERN", r"v\d+\.\d+\.\d+.*"
  )
  test_folder = os.getenv(
      "CPT_TEST_FOLDER", os.path.join("conan", "test_package")
  )
  upload_only_when_stable = os.getenv("CONAN_UPLOAD_ONLY_WHEN_STABLE", True)

  builder = ConanMultiPackager(
      reference=get_reference(username),
      username=username,
      login_username=login_username,
      upload=upload,
      stable_branch_pattern=stable_branch_pattern,
      upload_only_when_stable=upload_only_when_stable,
      test_folder=test_folder,
  )
  builder.add_common_builds(pure_c=False)
  builder.run()



# Copyright 2022 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from pathlib import Path
import platform
import subprocess

parser = argparse.ArgumentParser()
parser.add_argument(
    "--flatc",
    help="path of the Flat C compiler relative to the root directory",
)
parser.add_argument("--cpp-0x", action="store_true", help="use --cpp-std c++ox")
parser.add_argument(
    "--skip-monster-extra",
    action="store_true",
    help="skip generating tests involving monster_extra.fbs",
)
parser.add_argument(
    "--skip-gen-reflection",
    action="store_true",
    help="skip generating the reflection.fbs files",
)
args = parser.parse_args()

# Get the path where this script is located so we can invoke the script from
# any directory and have the paths work correctly.
script_path = Path(__file__).parent.resolve()

# Get the root path as an absolute path, so all derived paths are absolute.
root_path = script_path.parent.absolute()
tests_path = Path(root_path, "tests")

# Get the location of the flatc executable, reading from the first command line
# argument or defaulting to default names.
flatc_exe = Path(
    ("flatc" if not platform.system() == "Windows" else "flatc.exe")
    if not args.flatc
    else args.flatc
)

# Find and assert flatc compiler is present.
if root_path in flatc_exe.parents:
  flatc_exe = flatc_exe.relative_to(root_path)
flatc_path = Path(root_path, flatc_exe)
assert flatc_path.exists(), "Cannot find the flatc compiler " + str(flatc_path)


# Execute the flatc compiler with the specified parameters
def flatc(
    options, schema, prefix=None, include=None, data=None, cwd=tests_path
):
  cmd = [str(flatc_path)] + options
  if prefix:
    cmd += ["-o"] + [prefix]
  if include:
    cmd += ["-I"] + [include]
  cmd += [schema] if isinstance(schema, str) else schema
  if data:
    cmd += [data] if isinstance(data, str) else data
  result = subprocess.run(cmd, cwd=str(cwd), check=True)



# Copyright 2013-2025 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.

"""
Test low-level LMDB API (py-lmdb compatibility)

This test is based on the py-lmdb address-book.py example:
https://github.com/jnwatson/py-lmdb/blob/master/examples/address-book.py

It demonstrates that zlmdb provides a fully compatible py-lmdb API
for direct LMDB access.
"""

import tempfile
import shutil
import os

import zlmdb.lmdb as lmdb
import pytest


@pytest.fixture
def tmpdir():
    """Create a temporary directory for test database"""
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path)


def test_address_book_example(tmpdir):
    """
    Test the complete py-lmdb address-book example workflow.

    This verifies:
    - Environment creation with multiple databases
    - Subdatabase creation
    - Write transactions
    - Read transactions
    - Cursor iteration (sorted keys)
    - Update operations
    - Delete operations
    - Drop database operations
    """
    dbpath = os.path.join(tmpdir, "address-book.lmdb")

    # Open (and create if necessary) our database environment. Must specify
    # max_dbs=... since we're opening subdbs.
    env = lmdb.open(dbpath, max_dbs=10)

    # Now create subdbs for home and business addresses.
    home_db = env.open_db(b"home")
    business_db = env.open_db(b"business")

    # Add some telephone numbers to each DB:
    with env.begin(write=True) as txn:
        txn.put(b"mum", b"012345678", db=home_db)
        txn.put(b"dad", b"011232211", db=home_db)
        txn.put(b"dentist", b"044415121", db=home_db)
        txn.put(b"hospital", b"078126321", db=home_db)

        txn.put(b"vendor", b"0917465628", db=business_db)
        txn.put(b"customer", b"0553211232", db=business_db)
        txn.put(b"coworker", b"0147652935", db=business_db)
        txn.put(b"boss", b"0123151232", db=business_db)
        txn.put(b"manager", b"0644810485", db=business_db)

    # Verify: Iterate each DB to show the keys are sorted
    with env.begin() as txn:
        # Home database - check sorted order
        home_entries = list(txn.cursor(db=home_db))
        assert home_entries == [
            (b"dad", b"011232211"),
            (b"dentist", b"044415121"),
            (b"hospital", b"078126321"),
            (b"mum", b"012345678"),
        ]

        # Business database - check sorted order
        business_entries = list(txn.cursor(db=business_db))
        assert business_entries == [
            (b"boss", b"0123151232"),
            (b"coworker", b"0147652935"),
            (b"customer", b"0553211232"),
            (b"manager", b"0644810485"),
            (b"vendor", b"0917465628"),
        ]

    # Now let's update some phone numbers. We can specify the default subdb when
    # starting the transaction, rather than pass it in every time:
    with env.begin(write=True, db=home_db) as txn:
        # Update dentist number
        txn.put(b"dentist", b"099991231")

        # Delete hospital number
        txn.delete(b"hospital")

        # Verify home DB state
        home_entries = list(txn.cursor())
        assert home_entries == [
            (b"dad", b"011232211"),
            (b"dentist", b"099991231"),  # Updated
            (b"mum", b"012345678"),
        ]
        # Note: hospital should be deleted
        assert b"hospital" not in [key for key, _ in home_entries]

    # Now let's look up a number in the business DB
    with env.begin(db=business_db) as txn:
        boss_number = txn.get(b"boss")
        assert boss_number == b"0123151232"

    # We got fired, time to delete all keys from the business DB.
    with env.begin(write=True) as txn:
        txn.drop(business_db, delete=False)

        # Add number for recruiter to business DB
        txn.put(b"recruiter", b"04123125324", db=business_db)

        # Verify business DB is now only the recruiter
        business_entries = list(txn.cursor(db=business_db))
        assert business_entries == [
            (b"recruiter", b"04123125324"),
        ]

    # Clean up
    env.close()


def test_lmdb_basic_operations(tmpdir):
    """
    Test basic LMDB operations through py-lmdb compatible API.

    This is a simpler test focusing on core CRUD operations.
    """
    dbpath = os.path.join(tmpdir, "test.lmdb")

    # Create environment and database
    env = lmdb.open(dbpath)

    # Write data
    with env.begin(write=True) as txn:
        txn.put(b"key1", b"value1")
        txn.put(b"key2", b"value2")
        txn.put(b"key3", b"value3")

    # Read data
    with env.begin() as txn:
        assert txn.get(b"key1") == b"value1"
        assert txn.get(b"key2") == b"value2"
        assert txn.get(b"key3") == b"value3"
        assert txn.get(b"nonexistent") is None

    # Update data
    with env.begin(write=True) as txn:
        txn.put(b"key2", b"new_value2")

    # Verify update
    with env.begin() as txn:
        assert txn.get(b"key2") == b"new_value2"

    # Delete data
    with env.begin(write=True) as txn:
        assert txn.delete(b"key1") is True
        assert txn.delete(b"nonexistent") is False

    # Verify deletion
    with env.begin() as txn:
        assert txn.get(b"key1") is None
        assert txn.get(b"key2") == b"new_value2"
        assert txn.get(b"key3") == b"value3"

    env.close()


def test_lmdb_cursor_operations(tmpdir):
    """
    Test cursor operations for iteration and positioning.
    """
    dbpath = os.path.join(tmpdir, "cursor-test.lmdb")
    env = lmdb.open(dbpath)

    # Insert data with predictable sorting
    with env.begin(write=True) as txn:
        for i in range(10):
            key = f"key{i:02d}".encode()
            value = f"value{i}".encode()
            txn.put(key, value)

    # Test cursor iteration (should be sorted)
    with env.begin() as txn:
        cursor = txn.cursor()
        entries = list(cursor)

        # Verify sorted order
        assert len(entries) == 10
        assert entries[0] == (b"key00", b"value0")
        assert entries[9] == (b"key09", b"value9")

        # Verify all entries are in order
        for i, (key, value) in enumerate(entries):
            expected_key = f"key{i:02d}".encode()
            expected_value = f"value{i}".encode()
            assert key == expected_key
            assert value == expected_value

    # Test cursor positioning
    with env.begin() as txn:
        cursor = txn.cursor()

        # Move to first
        assert cursor.first() is True
        assert cursor.key() == b"key00"

        # Move to last
        assert cursor.last() is True
        assert cursor.key() == b"key09"

        # Move to specific key
        assert cursor.set_key(b"key05") is True
        assert cursor.key() == b"key05"
        assert cursor.value() == b"value5"

        # Move next
        assert cursor.next() is True
        assert cursor.key() == b"key06"

        # Move prev
        assert cursor.prev() is True
        assert cursor.key() == b"key05"

    env.close()


def test_lmdb_environment_info(tmpdir):
    """
    Test environment info and statistics.
    """
    dbpath = os.path.join(tmpdir, "info-test.lmdb")
    env = lmdb.open(dbpath, map_size=10 * 1024 * 1024)  # 10MB

    # Get environment info
    info = env.info()
    assert info["map_size"] == 10 * 1024 * 1024
    assert info["last_pgno"] >= 0
    assert info["last_txnid"] >= 0
    assert info["max_readers"] > 0
    assert info["num_readers"] >= 0

    # Get environment stats
    stats = env.stat()
    assert stats["psize"] > 0
    assert stats["depth"] >= 0
    assert stats["entries"] >= 0

    env.close()


if __name__ == "__main__":
    # Allow running as a script for quick testing
    import sys

    tmpdir = tempfile.mkdtemp()
    try:
        test_address_book_example(tmpdir)
        test_lmdb_basic_operations(tmpdir)
        test_lmdb_cursor_operations(tmpdir)
        test_lmdb_environment_info(tmpdir)
        print("✓ All tests passed!")
    finally:
        shutil.rmtree(tmpdir)



from golden_utils import flatc_golden


def flatc(options, schema):
  # Wrap the golden flatc generator with Java specifics
  flatc_golden(options=["--java"] + options, schema=schema, prefix="java")


def GenerateJava():
  flatc([], "basic.fbs")



import re


def version() -> tuple[int, int, int, int | None, str | None]:
    """
    Return the exact git version of a vendored Git repo, e.g. FlatBuffers runtime.

    git describe --tags --abbrev=0

    # Checkout the latest tag closest to the current branch tip
    git checkout $(git describe --tags --abbrev=0)

    Handles:

    1. "v25.9.23"              -> (25, 9, 23, None, None)       # Release (Named Tag, CalVer Year.Month.Day)
    2. "v25.9.23-71"           -> (25, 9, 23, 71, None)         # 71 commits ahead of the Release v25.9.23
    3. "v25.9.23-71-g19b2300f" -> (25, 9, 23, 71, "19b2300f")   # dito, with Git commit hash
    """

    # Pattern explanation:
    # ^v                : Start of string, literal 'v'
    # (\d+)\.(\d+)\.(\d+) : Groups 1,2,3 - Major.Minor.Patch (Required)
    #
    # (?: ... )?        : Non-capturing group (grouping only, not saved), optional '?'
    # -(\d+)            : Literal hyphen, Group 4 (Commits)
    #
    # (?: ... )?        : Non-capturing group, optional '?'
    # -g                : Literal hyphen and 'g' (separator)
    # ([0-9a-f]+)       : Group 5 (Hash)

    pattern = r"^v(\d+)\.(\d+)\.(\d+)(?:-(\d+))?(?:-g([0-9a-f]+))?$"

    match = re.match(pattern, __git_version__)

    if match:
        major, minor, patch, commits, commit_hash = match.groups()

        # Convert commits to int if present, else None
        commits_int = int(commits) if commits else None

        return (int(major), int(minor), int(patch), commits_int, commit_hash)

    # Fallback if regex fails entirely (returns 0.0.0 to satisfy type hint)
    return (0, 0, 0, None, None)



from twisted.internet.task import react
from twisted.internet.defer import inlineCallbacks

from autobahn.twisted import util

import txaio
from txaioetcd import Client, KeySet

import yaml

from pprint import pprint, pformat
import random

import zlmdb
from zlmdb._pmap import  MapStringString


try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory



class MySchema(zlmdb.Schema):

    samples: MapStringString

    def __init__(self):
        self.samples = MapStringString(1)


@inlineCallbacks
def main(reactor):
    if True:
        with TemporaryDirectory() as dbpath:
            print('Using temporary directory {} for database'.format(dbpath))

            schema = MySchema()

            with zlmdb.Database(dbpath) as db:
                # write records into zlmdb
                with db.begin(write=True) as txn:
                    for i in range(10):
                        key = 'key{}'.format(i)
                        value = 'value{}'.format(random.randint(0, 1000))
                        schema.samples[txn, key] = value

                # read records from zlmdb
                with db.begin() as txn:
                    for i in range(10):
                        key = 'key{}'.format(i)
                        value = schema.samples[txn, key]
                        print('key={} : value={}'.format(key, value))

    if True:
        # etcd database
        etcd = Client(reactor)
        status = yield etcd.status()
        print(status)

        # zlmdb database
        schema = MySchema()
        dbpath = '/tmp/.test-zlmdb'

        with zlmdb.Database(dbpath) as db:
            print('zlmdb open on {}'.format(dbpath))

            # check current record count
            with db.begin() as txn:
                cnt = schema.samples.count(txn)
                print('currently {} rows in table'.format(cnt))

            # watch changes in etcd and write to local zlmdb
            def on_change(kv):
                key = kv.key.decode()
                value = kv.value.decode()
                with db.begin(write=True) as txn:
                    schema.samples[txn, key] = value
                print('on_change received from etcd and written to zlmdb: key={} value={}'.format(key, value))

            # start watching for etcd changes ..
            ks = [KeySet('k'.encode(), prefix=True)]
            d = etcd.watch(ks, on_change)

            print('watching for 1s ..')
            yield txaio.sleep(1)

            # loop every 1s and write a key-value in etcd directly
            for i in range(5):
                print('watching for 1s ..')
                yield txaio.sleep(1)

                key = 'key{}'.format(i).encode()
                value = 'value{}'.format(random.randint(0, 1000)).encode()

                etcd.set(key, value)

            # cancel our watch
            d.cancel()

            yield util.sleep(1)

            # check current record count
            with db.begin() as txn:
                cnt = schema.samples.count(txn)
                print('currently {} rows in table'.format(cnt))

            yield util.sleep(1)


if __name__ == '__main__':
    txaio.start_logging(level='info')
    react(main)



import flatbuffers
from _gen.crossbarfx import User, Date, Tag, Rating

RESULT = {
    'objects': 0,
    'bytes': 0
}


def test():
    global RESULT

    builder = flatbuffers.Builder(0)

    name = builder.CreateString('Homer Simpson')
    authid = builder.CreateString('homer')
    email = builder.CreateString('homer.simpson@example.com')

    User.UserStartTagsVector(builder, 2)
    builder.PrependUOffsetTRelative(Tag.Tag.GEEK)
    builder.PrependUOffsetTRelative(Tag.Tag.VIP)
    tags = builder.EndVector(2)

    ratings = None
    if False:
        _ratings = {
            'dawn-of-the-dead': 6.9,
            'day-of-the-dead': 7.5,
            'land-of-the-dead': 8.9
        }

        _ratings_strings = {
        }
        for name in _ratings.keys():
            _name = builder.CreateString(name)
            _ratings_strings[_name] = name

        User.UserStartRatingsVector(builder, len(_ratings))
        l = []
        for _name, _rating in _ratings.items():
            Rating.RatingStart(builder)
            Rating.RatingAddName(builder, _ratings_strings[_name])
            Rating.RatingAddRating(builder, _rating)
            rating = Rating.RatingEnd(builder)
            l.append(rating)
        ratings = builder.EndVector(len(_ratings))

    User.UserStart(builder)

    User.UserAddName(builder, name)
    User.UserAddAuthid(builder, authid)
    User.UserAddEmail(builder, email)
    User.UserAddBirthday(builder, Date.CreateDate(builder, 1950, 12, 24))
    User.UserAddIsFriendly(builder, True)
    User.UserAddTags(builder, tags)

    if ratings:
        User.UserAddRatings(builder, ratings)

    user = User.UserEnd(builder)

    builder.Finish(user)

    buf = builder.Output()

    #data = bytes(buf)

    RESULT['objects'] += 1
    RESULT['bytes'] += len(buf)
#    RESULT['bytes'] += len(data)


import timeit

N = 1000
M = 100000

for i in range(N):
    secs = timeit.timeit(test, number=M)
    ops = round(float(M) / secs, 1)
    print('{} objects/sec'.format(ops))
    print(RESULT)




from twisted.internet.task import react
from twisted.internet.defer import inlineCallbacks

import txaio
from autobahn.twisted import util

import numpy as np
import pandas as pd
import pyarrow as pa

import zlmdb
from zlmdb._pmap import _StringKeysMixin, PersistentMap


try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory


class _DataFrameValuesMixin(object):

    def __init__(self, marshal=None, unmarshal=None):
        self._marshal = marshal or self._zlmdb_marshal
        self._unmarshal = unmarshal or self._zlmdb_unmarshal

    def _serialize_value(self, value):
        return pa.serialize(value).to_buffer()

    def _deserialize_value(self, data):
        return pa.deserialize(data)


class MapStringDataFrame(_StringKeysMixin, _DataFrameValuesMixin, PersistentMap):

    def __init__(self, slot=None, compress=None):
        PersistentMap.__init__(self, slot=slot, compress=compress)


class MySchema(zlmdb.Schema):

    samples: MapStringDataFrame

    def __init__(self):
        self.samples = MapStringDataFrame(1)


@inlineCallbacks
def main(reactor):

    with TemporaryDirectory() as dbpath:
        print('Using temporary directory {} for database'.format(dbpath))

        schema = MySchema()
        db = zlmdb.Database(dbpath)

        # WRITE some native pandas data frames to zlmdb
        with db.begin(write=True) as txn:
            for i in range(10):
                if i % 2:
                    key = 'key{}'.format(i)
                    value = pd.DataFrame(np.random.randn(8, 4),
                                         columns=['A','B','C','D'])
                    schema.samples[txn, key] = value

        # READ back native pandas data frames from zlmdb
        with db.begin() as txn:
            for i in range(10):
                key = 'key{}'.format(i)
                value = schema.samples[txn, key]
                print('key={} : value=\n{}'.format(key, value))

    yield util.sleep(1)


if __name__ == '__main__':
    txaio.start_logging(level='info')
    react(main)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Object(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Object()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsObject(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def ObjectBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x42\x46\x42\x53", size_prefixed=size_prefixed
        )

    # Object
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Object
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Object
    def Fields(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Field import Field

            obj = Field()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Object
    def FieldsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Object
    def FieldsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Object
    def IsStruct(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Object
    def Minalign(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Object
    def Bytesize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Object
    def Attributes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .KeyValue import KeyValue

            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Object
    def AttributesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Object
    def AttributesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Object
    def Documentation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # Object
    def DocumentationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Object
    def DocumentationIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # File that this Object is declared in.
    # Object
    def DeclarationFile(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def ObjectStart(builder):
    builder.StartObject(8)


def Start(builder):
    return ObjectStart(builder)


def ObjectAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0
    )


def AddName(builder, name):
    return ObjectAddName(builder, name)


def ObjectAddFields(builder, fields):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(fields), 0
    )


def AddFields(builder, fields):
    return ObjectAddFields(builder, fields)


def ObjectStartFieldsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartFieldsVector(builder, numElems):
    return ObjectStartFieldsVector(builder, numElems)


def ObjectAddIsStruct(builder, isStruct):
    builder.PrependBoolSlot(2, isStruct, 0)


def AddIsStruct(builder, isStruct):
    return ObjectAddIsStruct(builder, isStruct)


def ObjectAddMinalign(builder, minalign):
    builder.PrependInt32Slot(3, minalign, 0)


def AddMinalign(builder, minalign):
    return ObjectAddMinalign(builder, minalign)


def ObjectAddBytesize(builder, bytesize):
    builder.PrependInt32Slot(4, bytesize, 0)


def AddBytesize(builder, bytesize):
    return ObjectAddBytesize(builder, bytesize)


def ObjectAddAttributes(builder, attributes):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(attributes), 0
    )


def AddAttributes(builder, attributes):
    return ObjectAddAttributes(builder, attributes)


def ObjectStartAttributesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartAttributesVector(builder, numElems):
    return ObjectStartAttributesVector(builder, numElems)


def ObjectAddDocumentation(builder, documentation):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(documentation), 0
    )


def AddDocumentation(builder, documentation):
    return ObjectAddDocumentation(builder, documentation)


def ObjectStartDocumentationVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartDocumentationVector(builder, numElems):
    return ObjectStartDocumentationVector(builder, numElems)


def ObjectAddDeclarationFile(builder, declarationFile):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(declarationFile), 0
    )


def AddDeclarationFile(builder, declarationFile):
    return ObjectAddDeclarationFile(builder, declarationFile)


def ObjectEnd(builder):
    return builder.EndObject()


def End(builder):
    return ObjectEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import sys
import os
import pytest
import logging

import txaio

txaio.use_twisted()

import zlmdb  # noqa

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory  # type:ignore

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

if sys.version_info >= (3, 6):
    from _schema_py3 import User, Schema2
else:
    from _schema_py2 import User, Schema2


@pytest.fixture(scope="module")
def testset1():
    users = []
    for i in range(1000):
        user = User.create_test_user(i)
        users.append(user)
    return users


def test_open1():
    with TemporaryDirectory() as dbpath:
        with zlmdb.Database(dbpath) as db:
            assert db.maxsize == 10485760
            assert db.is_sync
            assert db.is_open
            assert not db.is_readonly
            assert not db.is_writemap


def test_open2():
    with TemporaryDirectory() as dbpath:
        with zlmdb.Database(dbpath) as db:
            assert db.is_open
        with zlmdb.Database(dbpath) as db:
            assert db.is_open


def test_open3():
    with TemporaryDirectory() as dbpath:
        db = zlmdb.Database(dbpath)
        assert db.is_open
        try:
            zlmdb.Database(dbpath)
            assert False, (
                "opening same dbpath twice in same process did not throw an exception"
            )
        except RuntimeError as e:
            # RuntimeError: tried to open same dbpath "/tmp/tmpwc1dw5c8" twice within same process
            assert "twice within same process" in str(e), (
                "exception did not contain text we excepted: {}".format(e)
            )
        except Exception as e:
            assert False, "unexpected exception {} raised".format(e)


def test_open4():
    with TemporaryDirectory() as dbpath:
        db1 = zlmdb.Database.open(dbpath)
        assert db1.is_open

        db2 = zlmdb.Database.open(dbpath)
        assert db2.is_open

        assert db1 == db2


def test_transaction():
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        with zlmdb.Database(dbpath) as db:
            with db.begin() as txn:
                logging.info("transaction open {}".format(txn.id()))
            logging.info("transaction committed")
        logging.info("database closed")


def test_save_load():
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        schema = Schema2()

        user = User.create_test_user()

        with zlmdb.Database(dbpath) as db:
            with db.begin(write=True) as txn:
                schema.users[txn, user.oid] = user
                logging.info("user saved")

                _user = schema.users[txn, user.oid]
                assert _user
                assert user == _user
                logging.info("user loaded")

            logging.info("transaction committed")

        logging.info("database closed")


def test_save_load_many_1(testset1):
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        schema = Schema2()

        with zlmdb.Database(dbpath) as db:
            with db.begin(write=True) as txn:
                for user in testset1:
                    schema.users[txn, user.oid] = user

                cnt = schema.users.count(txn)
                logging.info("user saved: {}".format(cnt))
                assert cnt == len(testset1)

            with db.begin() as txn:
                cnt = schema.users.count(txn)
                assert cnt == len(testset1)

        with zlmdb.Database(dbpath) as db:
            with db.begin() as txn:
                cnt = schema.users.count(txn)
                assert cnt == len(testset1)


def test_save_load_many_2(testset1):
    with TemporaryDirectory() as dbpath:
        logging.info("Using temporary directory {} for database".format(dbpath))

        schema = Schema2()

        oids = []

        with zlmdb.Database(dbpath) as db:
            # write records in a 1st transaction
            with db.begin(write=True) as txn:
                c = 0
                for user in testset1:
                    schema.users[txn, user.oid] = user
                    oids.append(user.oid)
                    c += 1
                assert c == len(testset1)
                logging.info("[1] successfully stored {} records".format(c))

                # in the same transaction, read back records
                c = 0
                for oid in oids:
                    user = schema.users[txn, oid]
                    if user:
                        c += 1
                assert c == len(testset1)
                logging.info("[1] successfully loaded {} records".format(c))

            # in a new transaction, read back records
            c = 0
            with db.begin() as txn:
                for oid in oids:
                    user = schema.users[txn, oid]
                    if user:
                        c += 1
            assert c == len(testset1)
            logging.info("[2] successfully loaded {} records".format(c))

        # in a new database environment (and hence new transaction), read back records
        with zlmdb.Database(dbpath) as db:
            with db.begin() as txn:
                count = schema.users.count(txn)
                assert count == len(testset1)
                logging.info("total records: {}".format(count))

                c = 0
                for oid in oids:
                    user = schema.users[txn, oid]
                    if user:
                        c += 1
                assert c == len(testset1)
                logging.info("[3] successfully loaded {} records".format(c))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Conan recipe package for Google FlatBuffers"""
import os
import shutil
from conans import CMake, ConanFile, tools


class FlatbuffersConan(ConanFile):
  name = "flatbuffers"
  license = "Apache-2.0"
  url = "https://github.com/google/flatbuffers"
  homepage = "http://google.github.io/flatbuffers/"
  author = "Wouter van Oortmerssen"
  topics = ("conan", "flatbuffers", "serialization", "rpc", "json-parser")
  description = "Memory Efficient Serialization Library"
  settings = "os", "compiler", "build_type", "arch"
  options = {"shared": [True, False], "fPIC": [True, False]}
  default_options = {"shared": False, "fPIC": True}
  generators = "cmake"
  exports = "LICENSE"
  exports_sources = [
      "CMake/*",
      "include/*",
      "src/*",
      "grpc/*",
      "CMakeLists.txt",
      "conan/CMakeLists.txt",
  ]

  def source(self):
    """Wrap the original CMake file to call conan_basic_setup"""
    shutil.move("CMakeLists.txt", "CMakeListsOriginal.txt")
    shutil.move(os.path.join("conan", "CMakeLists.txt"), "CMakeLists.txt")

  def config_options(self):
    """Remove fPIC option on Windows platform"""
    if self.settings.os == "Windows":
      self.options.remove("fPIC")

  def configure_cmake(self):
    """Create CMake instance and execute configure step"""
    cmake = CMake(self)
    cmake.definitions["FLATBUFFERS_BUILD_TESTS"] = False
    cmake.definitions["FLATBUFFERS_BUILD_SHAREDLIB"] = self.options.shared
    cmake.definitions["FLATBUFFERS_BUILD_FLATLIB"] = not self.options.shared
    cmake.configure()
    return cmake

  def build(self):
    """Configure, build and install FlatBuffers using CMake."""
    cmake = self.configure_cmake()
    cmake.build()

  def package(self):
    """Copy Flatbuffers' artifacts to package folder"""
    cmake = self.configure_cmake()
    cmake.install()
    self.copy(pattern="LICENSE", dst="licenses")
    self.copy(
        pattern="FindFlatBuffers.cmake",
        dst=os.path.join("lib", "cmake", "flatbuffers"),
        src="CMake",
    )
    self.copy(pattern="flathash*", dst="bin", src="bin")
    self.copy(pattern="flatc*", dst="bin", src="bin")
    if self.settings.os == "Windows" and self.options.shared:
      if self.settings.compiler == "Visual Studio":
        shutil.move(
            os.path.join(self.package_folder, "lib", "%s.dll" % self.name),
            os.path.join(self.package_folder, "bin", "%s.dll" % self.name),
        )
      elif self.settings.compiler == "gcc":
        shutil.move(
            os.path.join(self.package_folder, "lib", "lib%s.dll" % self.name),
            os.path.join(self.package_folder, "bin", "lib%s.dll" % self.name),
        )

  def package_info(self):
    """Collect built libraries names and solve flatc path."""
    self.cpp_info.libs = tools.collect_libs(self)
    self.user_info.flatc = os.path.join(self.package_folder, "bin", "flatc")



from golden_utils import flatc_golden


def flatc(options, schema):
  # Wrap the golden flatc generator with PHP specifics
  flatc_golden(options=["--php"] + options, schema=schema, prefix="php")


def GeneratePhp():
  flatc([], "basic.fbs")



# Auto-generated by build_lmdb.py
# DO NOT EDIT - regenerated on each build

CONFIG = {
    'extra_compile_args': ['-UNDEBUG', '-DHAVE_PATCHED_LMDB=1', '-w'],
    'extra_sources': ['build/lmdb-src/mdb.c', 'build/lmdb-src/midl.c'],
    'extra_library_dirs': [],
    'extra_include_dirs': ['build/lmdb-src', 'lmdb-patches'],
    'libraries': [],
}



from golden_utils import flatc_golden


def flatc(options, schema):
  # Wrap the golden flatc generator with Swift specifics
  flatc_golden(options=["--ts"] + options, schema=schema, prefix="ts")


def GenerateTs():
  flatc([], "basic.fbs")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

# New schema language features that are not supported by old code generators.
class AdvancedFeatures(object):
    AdvancedArrayFeatures = 1
    AdvancedUnionFeatures = 2
    OptionalScalars = 4
    DefaultVectorsAndStrings = 8
